// Allows interaction with the server to track install rate
// and log messages.

'use strict';

/* For ESLint: List any global identifiers used in this file below */
/* global browser, log, getSettings, SURVEY, determineUserLanguage,
   blockCounts, getSubscriptionsMinusText, LocalCDN, recordGeneralMessage */

const STATS = (function statsConstuctor() {
  const userIDStorageKey = 'userid';
  const totalPingStorageKey = 'totalPings';
  const nextPingTimeStorageKey = 'next_ping_time';

  const statsUrl = 'https://ping.getadblock.com/stats/';

  const FiftyFiveMinutes = 3300000;

  let dataCorrupt = false;

  // Get some information about the version, os, and browser
  const { version } = browser.runtime.getManifest();
  let match = navigator.userAgent.match(/(CrOS \w+|Windows NT|Mac OS X|Linux|Android) ([\d._]+)?/);
  const os = (match || [])[1] || 'Unknown';
  const osVersion = (match || [])[2] || 'Unknown';
  const flavor = 'F';
  match = navigator.userAgent.match(/(?:Firefox)\/([\d.]+)/);
  const browserVersion = (match || [])[1] || 'Unknown';

  const firstRun = false;

  let userID;

  // Inputs: key:string.
  // Returns value if key exists, else undefined.
  const storageGetStats = function (key) {
    const store = localStorage;
    if (store === undefined) {
      return undefined;
    }

    const json = store.getItem(key);
    if (json == null) {
      return undefined;
    }
    try {
      return JSON.parse(json);
    } catch (e) {
      log(`Couldn't parse json for ${key}`);
      return undefined;
    }
  };

  // Inputs: key:string, value:object.
  // Note: this is stats specific storageSetStats that allows us
  // to capture any exception, and set the dataCorruption indicator
  // If value === undefined, removes key from storage.
  // Returns undefined.
  const storageSetStats = function (key, value) {
    const store = localStorage;
    if (value === undefined) {
      store.removeItem(key);
      return;
    }

    try {
      store.setItem(key, JSON.stringify(value));
    } catch (ex) {
      dataCorrupt = true;
    }
  };

  const chromeStorageSet = function (key, value, callback) {
    if (value === undefined) {
      browser.storage.local.removeItem(key);
      return;
    }

    const saveData = {};
    saveData[key] = value;
    browser.storage.local.set(saveData, callback);
  };

  const determineTotalPingCount = function (response) {
    let localTotalPings = storageGetStats(STATS.totalPingStorageKey);
    localTotalPings = Number.isNaN(localTotalPings) ? 0 : localTotalPings;
    let totalPings = response[STATS.totalPingStorageKey];
    totalPings = Number.isNaN(totalPings) ? 0 : totalPings;
    return Math.max(localTotalPings, totalPings);
  };

  // Give the user a userid if they don't have one yet.
  const checkUserId = function () {
    const userIDPromise = new Promise(((resolve) => {
      browser.storage.local.get(STATS.userIDStorageKey, (response) => {
        const localuserid = storageGetStats(STATS.userIDStorageKey);
        if (!response[STATS.userIDStorageKey] && !localuserid) {
          STATS.firstRun = true;
          const timeSuffix = (Date.now()) % 1e8; // 8 digits from end of
          // timestamp
          const alphabet = 'abcdefghijklmnopqrstuvwxyz0123456789';
          const result = [];
          for (let i = 0; i < 8; i += 1) {
            const choice = Math.floor(Math.random() * alphabet.length);
            result.push(alphabet[choice]);
          }

          userID = result.join('') + timeSuffix;
          // store in redudant locations
          chromeStorageSet(STATS.userIDStorageKey, userID);
          storageSetStats(STATS.userIDStorageKey, userID);
        } else {
          userID = response[STATS.userIDStorageKey] || localuserid;
          if (!response[STATS.userIDStorageKey] && localuserid) {
            chromeStorageSet(STATS.userIDStorageKey, userID);
          }

          if (response[STATS.userIDStorageKey] && !localuserid) {
            storageSetStats(STATS.userIDStorageKey, userID);
          }
        }

        resolve(userID);
      });
    }));

    return userIDPromise;
  };

  const getPingData = function (callbackFN) {
    if (!callbackFN && (typeof callbackFN !== 'function')) {
      return;
    }

    browser.storage.local.get(STATS.totalPingStorageKey, (response) => {
      const totalPings = determineTotalPingCount(response);
      const data = {
        u: userID,
        v: version,
        f: flavor,
        o: os,
        bv: browserVersion,
        ov: osVersion,
        ad: getSettings().show_advanced_options ? '1' : '0',
        l: determineUserLanguage(),
        st: SURVEY.types(),
        pc: totalPings,
        cb: '0',
        cdn: getSettings().local_cdn ? '1' : '0',
        cdnr: LocalCDN.getRedirectCount(),
        cdnd: LocalCDN.getDataCount(),
      };
      if (blockCounts
          && typeof blockCounts.get === 'function') {
        const bc = blockCounts.get();
        data.b = bc.total;
        data.mt = bc.malware_total;
      }

      if (browser.runtime.id) {
        data.extid = browser.runtime.id;
      }

      const subs = getSubscriptionsMinusText();
      if (subs && subs.acceptable_ads) {
        data.aa = subs.acceptable_ads.subscribed ? '1' : '0';
      } else {
        data.aa = 'u';
      }

      data.dc = dataCorrupt ? '1' : '0';
      callbackFN(data);
    });
  };

  const handlePingResponse = function (responseData) {
    SURVEY.maybeSurvey(responseData);
  };

  // Tell the server we exist.
  const pingNow = function () {
    getPingData((pingData) => {
      const data = pingData;
      if (!data.u) {
        return;
      }
      // attempt to stop users that are pinging us 'alot'
      // by checking the current ping count,
      // if the ping count is above a theshold,
      // then only ping 'occasionally'
      if (data.pc > 5000) {
        if (data.pc > 5000 && data.pc < 100000 && ((data.pc % 5000) !== 0)) {
          return;
        }

        if (data.pc >= 100000 && ((data.pc % 50000) !== 0)) {
          return;
        }
      }

      data.cmd = 'ping';
      const ajaxOptions = {
        jsonp: false,
        type: 'POST',
        url: statsUrl,
        data,
        success: handlePingResponse, // TODO: Remove when we no longer do a/b
        // tests
        error(e) {
          log('Ping returned error: ', e.status);
        },
      };
      if (browser.management && browser.management.getSelf) {
        browser.management.getSelf((info) => {
          data.it = info.installType.charAt(0);
          $.ajax(ajaxOptions);
        });
      } else {
        $.ajax(ajaxOptions);
      }

      // send Local CDN missed versions stats as well
      recordGeneralMessage('cdn_miss_stats', undefined, { cdnm: LocalCDN.getMissedVersions() });
    });
  };


  // Called just after we ping the server, to schedule our next ping.
  // Called just after we ping the server, to schedule our next ping.
  const scheduleNextPing = function () {
    browser.storage.local.get(STATS.totalPingStorageKey, (response) => {
      let totalPings = determineTotalPingCount(response);
      totalPings += 1;
      // store in redudant locations
      chromeStorageSet(STATS.totalPingStorageKey, totalPings, () => {
        if (browser.runtime.lastError) {
          dataCorrupt = true;
        }
      });

      storageSetStats(STATS.totalPingStorageKey, totalPings);

      let delayHours;

      if (totalPings === 1) {
        // Ping one hour after install
        delayHours = 1;
      } else if (totalPings < 9) {
        // Then every day for a week
        delayHours = 24;
      } else {
      // Then weekly forever
        delayHours = 24 * 7;
      }

      const millis = 1000 * 60 * 60 * delayHours;
      const nextPingTime = Date.now() + millis;

      // store in redudant location
      chromeStorageSet(STATS.nextPingTimeStorageKey, nextPingTime, () => {
        if (browser.runtime.lastError) {
          dataCorrupt = true;
        }
      });

      storageSetStats(STATS.nextPingTimeStorageKey, nextPingTime);
    });
  };

  // Return the number of milliseconds until the next scheduled ping.
  const millisTillNextPing = function (callbackFN) {
    if (!callbackFN || (typeof callbackFN !== 'function')) {
      return;
    }
    // If we've detected data corruption issues,
    // then default to a 55 minute ping interval
    if (dataCorrupt) {
      callbackFN(FiftyFiveMinutes);
      return;
    }
    // Wait 10 seconds to allow the previous 'set' to finish
    window.setTimeout(() => {
      browser.storage.local.get(STATS.nextPingTimeStorageKey, (response) => {
        let localNextPingTime = storageGetStats(STATS.nextPingTimeStorageKey);
        localNextPingTime = Number.isNaN(localNextPingTime) ? 0 : localNextPingTime;
        const storedPingTime = response[STATS.nextPingTimeStorageKey];
        let nextPingTime = Number.isNaN(storedPingTime) ? 0 : storedPingTime;
        nextPingTime = Math.max(localNextPingTime, nextPingTime);

        // if this is the first time we've run (just installed), millisTillNextPing is 0
        if (nextPingTime === 0 && STATS.firstRun) {
          callbackFN(0);
          return;
        }
        // if we don't have a 'next ping time', or it's not a valid number,
        // default to 55 minute ping interval
        if (nextPingTime === 0 || Number.isNaN(nextPingTime)) {
          callbackFN(FiftyFiveMinutes);
          return;
        }

        callbackFN(nextPingTime - Date.now());
      }); // end of get
    }, 10000);
  };

  // Used to rate limit .message()s.  Rate limits reset at startup.
  const throttle = {
    // A small initial amount in case the server is bogged down.
    // The server will tell us the correct amount.
    maxEventsPerHour: 3, // null if no limit
    // Called when attempting an event. If not rate limited, returns
    // true and records the event.
    attempt() {
      const now = Date.now();
      const oneHour = 1000 * 60 * 60;
      const times = this.event_times;
      const mph = this.maxEventsPerHour;
      // Discard old or irrelevant events
      while (times[0] && (times[0] + oneHour < now || mph === null)) {
        times.shift();
      }

      if (mph === null) {
        return true; // no limit
      }

      if (times.length >= mph) {
        return false; // used our quota this hour
      }

      times.push(now);
      return true;
    },

    event_times: [],
  };

  return {
    userIDStorageKey,
    totalPingStorageKey,
    nextPingTimeStorageKey,
    // True if AdBlock was just installed.
    firstRun,
    checkUserId,
    userId() {
      return userID;
    },
    version,
    flavor,
    browser: 'Firefox',
    browserVersion,
    os,
    osVersion,
    statsUrl,
    // Ping the server when necessary.
    startPinging() {
      function sleepThenPing() {
        millisTillNextPing((delay) => {
          window.setTimeout(() => {
            pingNow();
            scheduleNextPing();
            sleepThenPing();
          }, delay);
        });
      }

      checkUserId();

      // This will sleep, then ping, then schedule a new ping, then
      // call itself to start the process over again.
      sleepThenPing();
    },

    // Record some data, if we are not rate limited.
    msg(message) {
      if (!throttle.attempt()) {
        log('Rate limited:', message);
        return;
      }

      const data = {
        cmd: 'msg2',
        m: message,
        u: userID,
        v: version,
        fr: firstRun,
        f: flavor,
        bv: browserVersion,
        o: os,
        ov: osVersion,
      };
      if (browser.runtime.id) {
        data.extid = browser.runtime.id;
      }

      $.ajax(statsUrl, {
        type: 'POST',
        data,
        jsonp: false,
        complete(xhr) {
          let mph = parseInt(xhr.getResponseHeader('X-RateLimit-MPH'), 10);
          if (Number.isNaN(mph) || mph < -1) {
            // Server is sick
            mph = 1;
          }
          if (mph === -1) {
            // no rate limit
            mph = null;
          }
          throttle.maxEventsPerHour = mph;
        },
      });
    },
  };
}());
